# coding: utf8
import base64
import os
import passlib.context
import pytest
import sqlalchemy
import ws.nginxdbauth.web


@pytest.fixture
def database(tmpdir):
    db = sqlalchemy.create_engine('sqlite:////%s' % tmpdir.join('auth.db'))
    db.execute(
        'CREATE TABLE users ('
        'id integer, username varchar(40), password varchar(40), '
        'role varchar(40))')
    db.execute('INSERT INTO users VALUES (1, "normal", "asdf", "normal")')
    db.execute('INSERT INTO users VALUES (2, "super", "qwer", "super")')
    return db


@pytest.fixture
def config(monkeypatch):
    os.environ['NGINXDBAUTH_CONFIG'] = '/dev/null'

    def inner(config):
        monkeypatch.setattr('ws.nginxdbauth.web.ConfigParser.get',
                            lambda self, s, x: config.get(x))
        monkeypatch.setattr('ws.nginxdbauth.web.ConfigParser.has_option',
                            lambda self, s, x: x in config)
    return inner


def basic_auth(username, password):
    return {'Authorization': 'Basic '.encode('ascii') + base64.b64encode(
        u':'.join([username, password]).encode('latin1'))}


def test_correct_password_returns_200(config, database):
    config({'dsn': database.url, 'query': 'SELECT id FROM users WHERE '
            'username = :username AND password = :password'})
    b = ws.nginxdbauth.web.app.test_client()
    r = b.get('/', headers=basic_auth('normal', 'asdf'))
    assert r.status_code == 200, r.data.decode('ascii')


def test_wrong_password_returns_403(config, database):
    config({'dsn': database.url, 'query': 'SELECT id FROM users WHERE '
            'username = :username AND password = :password'})
    b = ws.nginxdbauth.web.app.test_client()
    r = b.get('/', headers=basic_auth('normal', 'invalid'))
    assert r.status_code == 403, r.data.decode('ascii')


def test_headers_are_available_as_query_parameters(config, database):
    config({
        'dsn': database.url, 'query': 'SELECT id FROM users WHERE '
        'username = :username AND password = :password AND role = :x_role'})
    b = ws.nginxdbauth.web.app.test_client()
    headers = basic_auth('super', 'qwer')
    headers['X-Role'] = 'super'
    r = b.get('/', headers=headers)
    assert r.status_code == 200, r.data.decode('ascii')

    headers = basic_auth('normal', 'asdf')
    headers['X-Role'] = 'super'
    r = b.get('/', headers=headers)
    assert r.status_code == 403, r.data.decode('ascii')


def test_handles_non_ascii_entries(config, database):
    database.execute(u'INSERT INTO users VALUES (3, "ümläut", "asdf", "")')
    config({'dsn': database.url, 'query': 'SELECT id FROM users WHERE '
            'username = :username AND password = :password'})
    b = ws.nginxdbauth.web.app.test_client()
    r = b.get('/', headers=basic_auth(u'ümläut', 'asdf'))
    assert r.status_code == 200, r.data.decode('ascii')


def test_password_hashing_with_passlib(config, database):
    pwd_context = passlib.context.CryptContext(schemes=['sha256_crypt'])
    database.execute('INSERT INTO users VALUES (3, "foo", :pw, "")',
                     pw=pwd_context.hash('secret'))
    config({'dsn': database.url,
            'query': 'SELECT password FROM users WHERE username = :username',
            'password_hash': 'sha256_crypt'})
    b = ws.nginxdbauth.web.app.test_client()
    r = b.get('/', headers=basic_auth('foo', 'secret'))
    assert r.status_code == 200, r.data.decode('ascii')


def test_no_auth_header_returns_401(config):
    b = ws.nginxdbauth.web.app.test_client()
    r = b.get('/')
    assert r.status_code == 401, r.data.encode('ascii')
    r = b.get('/', headers={'WWW-Authenticate': 'my realm'})
    assert r.status_code == 401, r.data.encode('ascii')
    assert r.headers['WWW-Authenticate'] == 'my realm'



from flask import Flask, request, make_response
import argparse
import logging
import os
import os.path
import sqlalchemy
import sys
import wsgiref.handlers
import wsgiref.simple_server

try:
    from ConfigParser import ConfigParser
except ImportError:
    from configparser import ConfigParser

app = Flask(__name__)
log = logging.getLogger(__name__)
LOG_FORMAT = '%(asctime)s %(levelname)-5.5s [%(name)s] %(message)s'


@app.route('/')
def auth_view():
    if not request.authorization:
        response = make_response('AUTHENTICATE', 401)
        if 'WWW-Authenticate' in request.headers:
            response.headers['WWW-Authenticate'] = request.headers[
                'WWW-Authenticate']
        return response
    config = ConfigParser()
    config.read(os.path.expanduser(os.environ['NGINXDBAUTH_CONFIG']))
    get = (lambda x: config.get('default', x)
           if config.has_option('default', x) else None)  # noqa
    db = sqlalchemy.create_engine(get('dsn'))
    params = {
        'username': request.authorization.username,
        'password': request.authorization.password,
    }
    if sys.version_info < (3,):
        # XXX Werkzeug bug, authorization properties should be unicode already.
        for key, value in list(params.items()):
            params[key] = value.decode('latin1')
    for key, value in request.headers:
        params[key.lower().replace('-', '_')] = value

    verified = False
    result = db.execute(sqlalchemy.text(get('query')), **params).fetchall()
    if len(result) == 1:
        hashing = get('password_hash')
        if hashing:
            import passlib  # soft dependency
            pwd_context = passlib.context.CryptContext(schemes=[hashing])
            verified = pwd_context.verify(params['password'], result[0][0])
        else:
            verified = True
    if verified:
        return 'OK', 200
    else:
        return 'FAIL', 403


@app.errorhandler(Exception)
def handle_error(error):
    log.error('An error occured', exc_info=True)
    return str(error), 500


def cgi():
    # We only have the one route
    os.environ['PATH_INFO'] = '/'
    logfile = os.environ.get('NGINXDBAUTH_LOGFILE')
    if logfile:
        logging.basicConfig(filename=logfile, format=LOG_FORMAT)
    wsgiref.handlers.CGIHandler().run(app.wsgi_app)


def serve():
    parser = argparse.ArgumentParser()
    parser.add_argument('--host', default='localhost', help='bind host')
    parser.add_argument('--port', default='8899', help='bind port', type=int)
    parser.add_argument('--config', help='path to config file')
    options = parser.parse_args()
    if options.config:  # auth_view will raise KeyError to signal missing param
        os.environ['NGINXDBAUTH_CONFIG'] = options.config
    logging.basicConfig(stream=sys.stdout, format=LOG_FORMAT)
    wsgiref.simple_server.make_server(
        options.host, options.port, app.wsgi_app).serve_forever()
